package pairingfunction

import kotlinx.coroutines.*
import kotlin.collections.ArrayList


suspend fun main(args: Array<String>) {
    /*
    val aSequence = sequence {
        while (true)
            this.yield("a")
    }*/
    //println(aSequence.take(10000).joinToString(""))
    // println(getFromAStern(arrayListOf("a", "b", "c"), 120000))
    println(f(compute(120000, 8))) // Entscheiden für ein Wort, welches in M enthalten ist und genügend Zeit gewährt wird
    println("---------------------------------------------")
    println(f(compute(120000, 3))) // Entscheiden für das gleiche Wort mit zu wenig Zeit

}

/**
 * Entscheidungsfunktion für die Teilmenge M. Sie enthält alle Wörter aus A*, die mit ab beginnen.
 *
 * @param w Wort, welches überprüft werden soll
 *
 * @return true, wenn w element von A* ansonsten dauert die Entscheidung endlos
 */
fun chiStrich(w: String): Any {
    Thread.sleep(500L) // erzwungene Verlängerung der Funktion
    return if (w.substring(0, 2) == "ab" && w.subSequence(2, w.length).all {
            it == 'a' || it == 'b' || it == 'c'
        }) {
        true
    } else
        eternity()
}

/**
 * Hilfsfunktion, die eine Endlosschleife erzeugt, um die Semi-Entscheidbarkeit zu simulieren
 */
tailrec fun eternity() {
    return eternity()
}

/**
 * Funktion, die die Entscheidung trifft, ob ein bestimmtes Wort aus A* zur Teilmenge M gehört.
 * Dazu wird ein codiertes Kantor-Paar verwendet, welche entschlüsselt wird. Die Berechnung wird nach der gegebenen Zeit abgebrochen.
 *
 * @param z codiertes Kantor-Paar
 *
 * @return das Wort aus A*, wenn Wort Teilmenge von M ist und die gewährte Zeit ausgereicht hat, ansonsten wird "aba" zurückgegeben
 */
suspend fun f(z: Long): String {
    val alphabet = arrayListOf<String>("a", "b", "c")
    val x = computeX(z)
    println("x: $x")
    val y = computeY(z)
    println("y: $y")
    var w: String = ""
    var istInM: Any? = null
    println("Bilde Wort aus A*...")
    w = getFromAStern(alphabet, x.toInt())
    println("Entscheide, ob Wort in M ist... (M Teilmenge von A* beginnend mit 'ab')")
    val job = GlobalScope.launch(Dispatchers.Default) {
        while (isActive) {
            istInM = chiStrich(w)
        }

    }
    delay(y * 100L)
    job.cancel()
    if (istInM != null) {
        return w
    } else {

        return "aba"
    }
}

/**
 * Berechnung der Codierung eines Cantor-Paares
 *
 * @param x Element, in diesem Fall das Wort aus A*
 * @param y Zeit, die gewährt wird, um die Berechnung durchzuführen
 *
 * @return Codierung des Paares als Long
 */
fun compute(x: Long, y: Long): Long {
    return (((x + y) * (x + y + 1)) / 2) + y
}

/**
 * Entschlüsseln der Codierung eines Cantor-Paares
 *
 * @param z Codierung enes Cantor-Paares als Long
 *
 * @return Element des Paares, in diesem Fall Wort aus A*
 */
fun computeX(z: Long): Long {
    val w: Long = Math.floor((Math.sqrt(8 * z.toDouble() + 1) - 1) / 2).toLong()
    return w - (z - w * (w + 1) / 2)
}
/**
 * Entschlüsseln der Codierung eines Cantor-Paares
 *
 * @param z Codierung enes Cantor-Paares als Long
 *
 * @return Zeit, die für die Berechnung gewährt wird
 */
fun computeY(z: Long): Long {
    val w: Long = Math.floor((Math.sqrt(8 * z.toDouble() + 1) - 1) / 2).toLong()
    return z - w * (w + 1) / 2
}

/**
* Funktion zum Erstellen von der Wortmenge A* bis zur angegebenen Länge. Die Menge ist längenlexikografisch geordnet.
 *
 * @param chars Buchstaben des Alphabets als Liste
 * @param length Länge der Wörter, bis zu der die Menge erzeugt werden soll
 *
 * @return Wortmenge A* als Liste
 */
fun aStern(chars: List<String>, length: Int): ArrayList<String> {
    val result = arrayListOf<String>()

    fun helper(prefix: String, chars: List<String>) {
        for (i in 0 until chars.size) {
            val word = prefix + chars[i]
            if (word.length == length) {
                result.add(prefix + chars[i])
            }
            if (word.length > length) {
                return
            }
            helper(prefix + chars[i], chars)
        }
    }
    helper("", chars)
    return result
}



/**
* Funktion zum Beziehen eines Wortes aus A* an dem übergeben index
*
* @param chars Buchstaben des Alphabets als ArrayList
*  @param index Index des gewünschten Wortes als Int
*/
fun getFromAStern(chars: ArrayList<String>, index: Int): String {
    if (index == 0) {
        return ""
    }
    if (index == 1) {
        return chars[0]
    }

    val length = Math.ceil(Math.log(((index / 2) + 1) / Math.log(2.0))).toInt()
    val m = arrayListOf<String>()
    for (i in 0..length + 1) {
        m.addAll(aStern(chars = chars, length = i))
    }

    return m[index - 1]
}

